from typing import (Optional,
                    Sequence)

from ground.hints import Point
from ground.linear import (SegmentsRelationship,
                           to_connected_segments_intersector,
                           to_segments_relater)
from prioq.base import PriorityQueue
from reprit.base import generate_repr

from .event import Event
from .utils import merge_ids


class EventsQueueKey:
    __slots__ = 'event',

    def __init__(self, event: Event) -> None:
        self.event = event

    __repr__ = generate_repr(__init__)

    def __lt__(self, other: 'EventsQueueKey') -> bool:
        """
        Checks if the event should be processed before the other.
        """
        event, other_event = self.event, other.event
        start_x, start_y = event.start.x, event.start.y
        other_start_x, other_start_y = other_event.start.x, other_event.start.y
        if start_x != other_start_x:
            # different x-coordinate,
            # the event with lower x-coordinate is processed first
            return start_x < other_start_x
        elif start_y != other_start_y:
            # different starts, but same x-coordinate,
            # the event with lower y-coordinate is processed first
            return start_y < other_start_y
        elif event.is_left_endpoint is not other_event.is_left_endpoint:
            # same start, but one is a left endpoint
            # and the other is a right endpoint,
            # the right endpoint is processed first
            return not event.is_left_endpoint
        else:
            # same start,
            # both events are left endpoints or both are right endpoints
            return event.end < other_event.end


class EventsQueue:
    __slots__ = '_queue', '_segments_relater', '_segments_intersector'

    def __init__(self) -> None:
        self._queue = PriorityQueue(key=EventsQueueKey)
        self._segments_relater = to_segments_relater()
        self._segments_intersector = to_connected_segments_intersector()

    __repr__ = generate_repr(__init__)

    def __bool__(self) -> bool:
        return bool(self._queue)

    def detect_intersection(self, below_event: Event, event: Event) -> None:
        relationship = self._segments_relater(
                below_event.start, below_event.end, event.start, event.end)
        if relationship is SegmentsRelationship.OVERLAP:
            # segments overlap
            starts_equal = event.start == below_event.start
            if starts_equal:
                start_min = start_max = None
            elif EventsQueueKey(event) < EventsQueueKey(below_event):
                start_min, start_max = event, below_event
            else:
                start_min, start_max = below_event, event
            ends_equal = event.end == below_event.end
            if ends_equal:
                end_min = end_max = None
            elif (EventsQueueKey(event.complement)
                  < EventsQueueKey(below_event.complement)):
                end_min, end_max = event.complement, below_event.complement
            else:
                end_min, end_max = below_event.complement, event.complement
            segments_ids = merge_ids(below_event.segments_ids,
                                     event.segments_ids)
            if starts_equal:
                if ends_equal:
                    # segments are equal
                    below_event.segments_ids = event.segments_ids = (
                        segments_ids)
                    event.set_both_relationships(relationship)
                    below_event.set_both_relationships(relationship)
                else:
                    # segments share the left endpoint
                    end_min.set_both_relationships(relationship)
                    end_max.complement.relationship = relationship
                    self._divide_segment(end_max.complement, end_min.start,
                                         segments_ids)
            elif ends_equal:
                # segments share the right endpoint
                start_max.set_both_relationships(relationship)
                start_min.complement.relationship = relationship
                self._divide_segment(start_min, start_max.start, segments_ids)
            elif start_min is end_max.complement:
                # one line segment includes the other one
                start_max.set_both_relationships(relationship)
                start_min_original_relationship = start_min.relationship
                start_min.relationship = relationship
                self._divide_segment(start_min, end_min.start, segments_ids)
                start_min.relationship = start_min_original_relationship
                start_min.complement.relationship = relationship
                self._divide_segment(start_min, start_max.start, segments_ids)
            else:
                # no line segment includes the other one
                start_max.relationship = relationship
                self._divide_segment(start_max, end_min.start, segments_ids)
                start_min.complement.relationship = relationship
                self._divide_segment(start_min, start_max.start, segments_ids)
        elif relationship is not SegmentsRelationship.NONE:
            # segments touch or cross
            point = self._segments_intersector(
                    below_event.start, below_event.end, event.start, event.end)
            if point != below_event.start and point != below_event.end:
                self._divide_segment(below_event, point)
            if point != event.start and point != event.end:
                self._divide_segment(event, point)
            event.set_both_relationships(max(event.relationship,
                                             relationship))
            below_event.set_both_relationships(max(below_event.relationship,
                                                   relationship))

    def peek(self) -> Event:
        return self._queue.peek()

    def pop(self) -> Event:
        return self._queue.pop()

    def push(self, event: Event) -> None:
        if event.start == event.end:
            raise ValueError('Degenerate segment found '
                             'with both endpoints being: {}.'
                             .format(event.start))
        self._queue.push(event)

    def _divide_segment(self,
                        event: Event,
                        break_point: Point,
                        segments_ids: Optional[Sequence[int]] = None) -> None:
        if segments_ids is None:
            segments_ids = event.segments_ids
        else:
            event.segments_ids = segments_ids
        left_event = Event(is_left_endpoint=True,
                           relationship=event.complement.relationship,
                           start=break_point,
                           complement=event.complement,
                           segments_ids=segments_ids)
        right_event = Event(is_left_endpoint=False,
                            relationship=event.relationship,
                            start=break_point,
                            complement=event,
                            segments_ids=segments_ids)
        event.complement.complement, event.complement = left_event, right_event
        self.push(left_event)
        self.push(right_event)



from typing import (Iterable,
                    List,
                    Sequence,
                    Tuple)

from ground.hints import (Point,
                          Segment)
from ground.linear import SegmentsRelationship

from .event import Event
from .events_queue import EventsQueue
from .sweep_line import SweepLine
from .utils import to_pairs_combinations


def sweep(segments: Sequence[Segment]) -> Iterable[Tuple[Event, Event]]:
    events_queue = to_events_queue(segments)
    sweep_line = SweepLine()
    prev_start = None
    prev_same_start_events = []  # type: List[Event]
    while events_queue:
        event = events_queue.pop()
        start = event.start
        same_start_events = (prev_same_start_events + [event]
                             if start == prev_start
                             else [event])
        while events_queue and events_queue.peek().start == start:
            same_start_events.append(events_queue.pop())
        for event, other_event in to_pairs_combinations(same_start_events):
            yield event, other_event
        for event in same_start_events:
            if event.is_left_endpoint:
                sweep_line.add(event)
                above_event, below_event = (sweep_line.above(event),
                                            sweep_line.below(event))
                if below_event is not None:
                    events_queue.detect_intersection(below_event, event)
                if above_event is not None:
                    events_queue.detect_intersection(event, above_event)
            else:
                event = event.complement
                if event in sweep_line:
                    above_event, below_event = (sweep_line.above(event),
                                                sweep_line.below(event))
                    sweep_line.remove(event)
                    if below_event is not None and above_event is not None:
                        events_queue.detect_intersection(below_event,
                                                         above_event)
                if len(event.segments_ids) > 1:
                    yield event, event
        prev_start, prev_same_start_events = start, same_start_events


def to_events_queue(segments: Sequence[Segment]) -> EventsQueue:
    endpoints_with_ids = sorted((sort_endpoints(segment), segment_id)
                                for segment_id, segment in enumerate(segments))
    result = EventsQueue()
    index = 0
    while index < len(endpoints_with_ids):
        endpoints, segment_id = endpoints_with_ids[index]
        index += 1
        same_segments_ids = [segment_id]
        while (index < len(endpoints_with_ids)
               and endpoints_with_ids[index][0] == endpoints):
            same_segments_ids.append(endpoints_with_ids[index][1])
            index += 1
        start, end = endpoints
        relationship = (SegmentsRelationship.NONE
                        if len(same_segments_ids) == 1
                        else SegmentsRelationship.OVERLAP)
        start_event = Event(is_left_endpoint=True,
                            relationship=relationship,
                            start=start,
                            complement=None,
                            segments_ids=same_segments_ids)
        end_event = Event(is_left_endpoint=False,
                          relationship=relationship,
                          start=end,
                          complement=start_event,
                          segments_ids=same_segments_ids)
        start_event.complement = end_event
        result.push(start_event)
        result.push(end_event)
    return result


def sort_endpoints(segment: Segment) -> Tuple[Point, Point]:
    start, end = segment.start, segment.end
    return (start, end) if start < end else (end, start)



from typing import (Dict,
                    Hashable,
                    Iterable,
                    Sequence,
                    Set,
                    Tuple)

from ground.geometries import to_segment_cls as _to_segment_cls
from ground.hints import (Contour,
                          Point,
                          Segment)
from ground.linear import (SegmentsRelationship as _SegmentsRelationship,
                           to_segments_intersector as _to_segments_intersector)

from .core.base import sweep as _sweep
from .core.utils import (merge_ids as _merge_ids,
                         to_pairs_combinations as _to_pairs_combinations)


def edges_intersect(contour: Contour) -> bool:
    """
    Checks if polygonal contour has self-intersection.

    Based on Shamos-Hoey algorithm.

    Time complexity:
        ``O(len(contour) * log len(contour))``
    Memory complexity:
        ``O(len(contour))``
    Reference:
        https://en.wikipedia.org/wiki/Sweep_line_algorithm

    :param contour: contour to check.
    :returns: true if contour is self-intersecting, false otherwise.

    .. note::
        Consecutive equal vertices like ``Point(2, 0)`` in

        .. code-block:: python

            Contour([Point(0, 0), Point(2, 0), Point(2, 0), Point(2, 2)])

        will be considered as self-intersection,
        if you don't want them to be treated as such
        -- filter out before passing as argument.

    >>> from ground.geometries import to_contour_cls, to_point_cls
    >>> Contour, Point = to_contour_cls(), to_point_cls()
    >>> edges_intersect(Contour([Point(0, 0), Point(2, 0), Point(2, 2)]))
    False
    >>> edges_intersect(Contour([Point(0, 0), Point(2, 0), Point(1, 0)]))
    True
    """
    vertices = contour.vertices
    if len(vertices) < 3:
        raise ValueError('Contour {contour} is degenerate.'
                         .format(contour=contour))
    if not _all_unique(vertices):
        return True
    segment_cls = _to_segment_cls()
    edges = [segment_cls(vertices[index - 1], vertices[index])
             for index in range(len(vertices))]

    def non_neighbours_intersect(edges_ids: Iterable[Tuple[int, int]],
                                 last_edge_index: int = len(edges) - 1
                                 ) -> bool:
        return any(next_segment_id - segment_id > 1
                   and (segment_id != 0 or next_segment_id != last_edge_index)
                   for segment_id, next_segment_id in edges_ids)

    return any((first_event.relationship is _SegmentsRelationship.OVERLAP
                or second_event.relationship is _SegmentsRelationship.OVERLAP
                or non_neighbours_intersect(_to_pairs_combinations(_merge_ids(
                    first_event.segments_ids, second_event.segments_ids))))
               for first_event, second_event in _sweep(edges))


def _all_unique(values: Iterable[Hashable]) -> bool:
    seen = set()
    seen_add = seen.add
    for value in values:
        if value in seen:
            return False
        else:
            seen_add(value)
    return True


def segments_intersect(segments: Sequence[Segment]) -> bool:
    """
    Checks if segments have at least one intersection.

    Based on Shamos-Hoey algorithm.

    Time complexity:
        ``O(len(segments) * log len(segments))``
    Memory complexity:
        ``O(len(segments))``
    Reference:
        https://en.wikipedia.org/wiki/Sweep_line_algorithm

    :param segments: sequence of segments.
    :returns: true if segments intersection found, false otherwise.

    >>> from ground.geometries import to_point_cls, to_segment_cls
    >>> Point, Segment = to_point_cls(), _to_segment_cls()
    >>> segments_intersect([])
    False
    >>> segments_intersect([Segment(Point(0, 0), Point(2, 2))])
    False
    >>> segments_intersect([Segment(Point(0, 0), Point(2, 0)),
    ...                     Segment(Point(0, 2), Point(2, 2))])
    False
    >>> segments_intersect([Segment(Point(0, 0), Point(2, 2)),
    ...                     Segment(Point(0, 0), Point(2, 2))])
    True
    >>> segments_intersect([Segment(Point(0, 0), Point(2, 2)),
    ...                     Segment(Point(2, 0), Point(0, 2))])
    True
    """
    return any(_sweep(segments))


def segments_cross_or_overlap(segments: Sequence[Segment]) -> bool:
    """
    Checks if at least one pair of segments crosses or overlaps.

    Based on Shamos-Hoey algorithm.

    Time complexity:
        ``O((len(segments) + len(intersections)) * log len(segments))``
    Memory complexity:
        ``O(len(segments))``
    Reference:
        https://en.wikipedia.org/wiki/Sweep_line_algorithm

    :param segments: sequence of segments.
    :returns: true if segments overlap or cross found, false otherwise.

    >>> from ground.geometries import to_point_cls, to_segment_cls
    >>> Point, Segment = to_point_cls(), _to_segment_cls()
    >>> segments_cross_or_overlap([])
    False
    >>> segments_cross_or_overlap([Segment(Point(0, 0), Point(2, 2))])
    False
    >>> segments_cross_or_overlap([Segment(Point(0, 0), Point(2, 0)),
    ...                            Segment(Point(0, 2), Point(2, 2))])
    False
    >>> segments_cross_or_overlap([Segment(Point(0, 0), Point(2, 2)),
    ...                            Segment(Point(0, 0), Point(2, 2))])
    True
    >>> segments_cross_or_overlap([Segment(Point(0, 0), Point(2, 2)),
    ...                            Segment(Point(2, 0), Point(0, 2))])
    True
    """
    relationships = _SegmentsRelationship.CROSS, _SegmentsRelationship.OVERLAP
    return any(first_event.relationship in relationships
               or second_event.relationship in relationships
               for first_event, second_event in _sweep(segments))


def segments_intersections(segments: Sequence[Segment]
                           ) -> Dict[Point, Set[Tuple[int, int]]]:
    """
    Returns mapping between intersection points
    and corresponding segments indices.

    Based on Bentley-Ottmann algorithm.

    Time complexity:
        ``O((len(segments) + len(intersections)) * log len(segments))``
    Memory complexity:
        ``O(len(segments) + len(intersections))``
    Reference:
        https://en.wikipedia.org/wiki/Bentley%E2%80%93Ottmann_algorithm

    >>> from ground.geometries import to_point_cls, to_segment_cls
    >>> Point, Segment = to_point_cls(), _to_segment_cls()
    >>> segments_intersections([]) == {}
    True
    >>> segments_intersections([Segment(Point(0, 0), Point(2, 2))]) == {}
    True
    >>> segments_intersections([Segment(Point(0, 0), Point(2, 0)),
    ...                         Segment(Point(0, 2), Point(2, 2))]) == {}
    True
    >>> (segments_intersections([Segment(Point(0, 0), Point(2, 2)),
    ...                          Segment(Point(0, 0), Point(2, 2))])
    ...  == {(0, 0): {(0, 1)}, (2, 2): {(0, 1)}})
    True
    >>> (segments_intersections([Segment(Point(0, 0), Point(2, 2)),
    ...                          Segment(Point(2, 0), Point(0, 2))])
    ...  == {(1, 1): {(0, 1)}})
    True

    :param segments: sequence of segments.
    :returns:
        mapping between intersection points and corresponding segments indices.
    """
    result = {}
    segments_intersector = _to_segments_intersector()
    for first_event, second_event in _sweep(segments):
        for segment_id, next_segment_id in _to_pairs_combinations(_merge_ids(
                first_event.segments_ids, second_event.segments_ids)):
            segment, next_segment = (segments[segment_id],
                                     segments[next_segment_id])
            for point in segments_intersector(segment.start, segment.end,
                                              next_segment.start,
                                              next_segment.end):
                result.setdefault(point, set()).add((segment_id,
                                                     next_segment_id))
    return result



from reprlib import recursive_repr
from typing import (Optional,
                    Sequence)

from ground.hints import Point
from ground.linear import SegmentsRelationship
from reprit.base import generate_repr


class Event:
    __slots__ = ('is_left_endpoint', 'relationship', 'start', 'complement',
                 'segments_ids')

    def __init__(self,
                 is_left_endpoint: bool,
                 relationship: SegmentsRelationship,
                 start: Point,
                 complement: Optional['Event'],
                 segments_ids: Sequence[int]) -> None:
        self.is_left_endpoint = is_left_endpoint
        self.relationship = relationship
        self.start = start
        self.complement = complement
        self.segments_ids = segments_ids

    __repr__ = recursive_repr()(generate_repr(__init__))

    @property
    def end(self) -> Point:
        return self.complement.start

    def set_both_relationships(self, relationship: SegmentsRelationship
                               ) -> None:
        self.relationship = self.complement.relationship = relationship



from functools import partial
from typing import (Callable,
                    Optional)

from dendroid import red_black
from ground.angular import (Orientation,
                            to_orienteer)
from ground.hints import Point
from reprit.base import generate_repr

from .event import Event
from .utils import merge_ids


class SweepLine:
    __slots__ = '_tree',

    def __init__(self) -> None:
        self._tree = red_black.set_(key=partial(SweepLineKey, to_orienteer()))

    __repr__ = generate_repr(__init__)

    def __contains__(self, event: Event) -> bool:
        return event in self._tree

    def add(self, event: Event) -> None:
        self._tree.add(event)

    def remove(self, event: Event) -> None:
        self._tree.remove(event)

    def above(self, event: Event) -> Optional[Event]:
        try:
            return self._tree.next(event)
        except ValueError:
            return None

    def below(self, event: Event) -> Optional[Event]:
        try:
            return self._tree.prev(event)
        except ValueError:
            return None


class SweepLineKey:
    __slots__ = 'orienteer', 'event'

    def __init__(self,
                 orienteer: Callable[[Point, Point, Point], Orientation],
                 event: Event) -> None:
        self.orienteer = orienteer
        self.event = event

    __repr__ = generate_repr(__init__)

    def __lt__(self, other: 'SweepLineKey') -> bool:
        """
        Checks if the segment (or at least the point) associated with event
        is lower than other's.
        """
        event, other_event = self.event, other.event
        if event is other_event:
            return False
        start, other_start = event.start, other_event.start
        end, other_end = event.end, other_event.end
        other_start_orientation = self.orienteer(end, start, other_start)
        other_end_orientation = self.orienteer(end, start, other_end)
        if other_start_orientation is other_end_orientation:
            start_x, start_y = start
            other_start_x, other_start_y = other_start
            if other_start_orientation is not Orientation.COLLINEAR:
                # other segment fully lies on one side
                return other_start_orientation is Orientation.COUNTERCLOCKWISE
            # segments are collinear
            elif start_x == other_start_x:
                end_x, end_y = end
                other_end_x, other_end_y = other_end
                if start_y != other_start_y:
                    # segments are vertical
                    return start_y < other_start_y
                # segments have same start
                elif end_y != other_end_y:
                    return end_y < other_end_y
                elif end_x != other_end_x:
                    # segments are horizontal
                    return end_x < other_end_x
                else:
                    # segments are equal
                    event.segments_ids = other_event.segments_ids = merge_ids(
                            event.segments_ids, other_event.segments_ids)
                    return False
            elif start_y != other_start_y:
                return start_y < other_start_y
            else:
                # segments are horizontal
                return start_x < other_start_x
        start_orientation = self.orienteer(other_end, other_start, start)
        end_orientation = self.orienteer(other_end, other_start, end)
        if start_orientation is end_orientation:
            return start_orientation is Orientation.CLOCKWISE
        elif other_start_orientation is Orientation.COLLINEAR:
            return other_end_orientation is Orientation.COUNTERCLOCKWISE
        elif start_orientation is Orientation.COLLINEAR:
            return end_orientation is Orientation.CLOCKWISE
        elif end_orientation is Orientation.COLLINEAR:
            return start_orientation is Orientation.CLOCKWISE
        else:
            return other_start_orientation is Orientation.COUNTERCLOCKWISE
